//
// Parago Media GmbH & Co. KG, Jrgen Burle (jbaurle@parago.de)
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED 
// TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
// CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
// DEALINGS IN THE SOFTWARE.
//

using System;
using System.Collections.Generic;
using System.Text;

namespace Parago.DALComp.Compiler {

	/// <summary>
	/// The Tokenizer class represents the first phase of the small compiler application
	/// and analyzes the source text by splitting it into tokens.
	/// </summary>
	public class Tokenizer {

		string text;
		Queue<Token> tokens;

		/// <summary>
		/// Initializes a new instance of the Tokenizer class when given a source text
		/// and starts lexing.
		/// </summary>
		public Tokenizer(string text) {
			// To avoid index overflow append new line character to text
			this.text=(text==null?String.Empty:text)+"\n";
			tokens=new Queue<Token>();
			Start();
		}

		/// <summary>
		/// Tokenize the source text and add the valid tokens to internal queue.
		/// </summary>
		void Start() {
			int i=0;
			StringBuilder value;
			while(i<text.Length) {
				switch(text[i]) {
					// Whitespace
					case '\n':
					case '\r':
					case '\t':
					case ' ':
						break;
					// Comment (until end of line)
					case '/':
						if(text[i+1]=='/')
							while(text[++i]!='\n') ;
						continue;
					case '(':
						tokens.Enqueue(new Token(TokenKind.ParenthesisOpen));
						break;
					case ')':
						tokens.Enqueue(new Token(TokenKind.ParenthesisClose));
						break;
					case '[':
						tokens.Enqueue(new Token(TokenKind.BracketOpen));
						break;
					case ']':
						tokens.Enqueue(new Token(TokenKind.BracketClose));
						break;
					case '{':
						tokens.Enqueue(new Token(TokenKind.CurlyBracketOpen));
						break;
					case '}':
						tokens.Enqueue(new Token(TokenKind.CurlyBracketClose));
						break;
					case '=':
						if(text[i+1]=='>') {
							i++;
							tokens.Enqueue(new Token(TokenKind.Assign));
						}
						else
							tokens.Enqueue(new Token(TokenKind.Equal));
						break;
					case ',':
						tokens.Enqueue(new Token(TokenKind.Comma));
						break;
				   // String (the combination '""' will be treated as one character and won't end the string)
					case '\"':
						value=new StringBuilder();
						do {
							i++;
							if(text[i]=='\r'||text[i]=='\n')
								Abort("Strings must be ended within the same line");
							if(text[i]=='\"') {
								if(text[i+1]!='\"') {  
									tokens.Enqueue(new Token(TokenKind.String, value.ToString()));
									break;
								}
								i++;
							}
							value.Append(text[i]);
						} while(i<text.Length);
						break;
					default:
						if(!Char.IsLetter(text[i]))
							Abort("Unrecognized token found");
						value=new StringBuilder(text[i].ToString());
						while(Char.IsLetterOrDigit(text[++i]))
							value.Append(text[i]);
						// Checking for keywords, types or identifiers
						switch(value.ToString().ToUpper()) {
							case "CONFIG":
								tokens.Enqueue(new Token(TokenKind.KeywordConfig));
								break;
							case "DALC":
								tokens.Enqueue(new Token(TokenKind.KeywordDALC));
								break;
							case "MAPPING":
								tokens.Enqueue(new Token(TokenKind.KeywordMapping));
								break;
							case "INT":
								tokens.Enqueue(new Token(TokenKind.Type, "System.Int32"));
								break;
							case "STRING":
								tokens.Enqueue(new Token(TokenKind.Type, "System.String"));
								break;
							// Must be an identifier
							default:
								tokens.Enqueue(new Token(TokenKind.Identifier, value.ToString()));
								break;
						}
						continue;
				}
				i++;
			}
		}

		/// <summary>
		/// Looks ahead the token stream adn returns the next token type without
		/// removing it from token queue.
		/// </summary>
		public TokenKind PeekTokenType() {
			// Always return at least TokenKind.EOT 
			return (tokens.Count>0)?tokens.Peek().Type:TokenKind.EOT;
		}

		/// <summary>
		/// Returns the next token and removes it from the token queue.
		/// </summary>
		public Token GetNextToken() {
			// Always return at least Token of type TokenKind.EOT
			return (tokens.Count>0)?tokens.Dequeue():new Token(TokenKind.EOT);
		}

		/// <summary>
		/// Aborts the lexing with specific message.
		/// </summary>
		void Abort(string message) {
			throw new Exception(String.Format("DALComp input error: {0}", message));
		}

	}

	/// <summary>
	/// Represents a single token ot type TokenKind.
	/// </summary>
	public class Token {

		public TokenKind Type;
		public string Value;

		public Token(TokenKind type) {
			Type=type;
			Value=null;
		}

		public Token(TokenKind type, string value) {
			Type=type;
			Value=value;
		}

	}

	/// <summary>
	/// Represents a enumeration of allowed token types.
	/// </summary>
	public enum TokenKind {
		KeywordConfig,
		KeywordDALC,
		KeywordMapping,
		Type,
		Identifier,
		String,
		Assign,					// =>
		Equal,					// =
		Comma,					// ,
		BracketOpen,			// [
		BracketClose,			// ]
		CurlyBracketOpen,		// {
		CurlyBracketClose,	// }
		ParenthesisOpen,		// (
		ParenthesisClose,		// )
		EOT						// End Of Text
	}

}
